/* -*- Mode: C; c-file-style: "bsd"; -*- */

#include "int.h"
#include "tree.h"
#include "mpoly.h"
#include "list.h"
#include "variable.h"
#include "op_bindings.h"
#include "errors.h"
#include "i18n.h"
#include "typecasts.h"

static NxpElement *
add(NxpOp op, NxpElement *a, NxpElement *b,
	NxpEvalContext *context, GError **error)
{
    NxpElement *pol;
    NxpElement *vector;
    int i, size, j;
    NxpElement *one, *zero, *one1;
    NxpElement *mon;

    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(a), NULL);

    j = nxp_element_list_contains(NXP_ELEMENT(NXP_ELEMENT_MPOLY(a)->args), b, context, error);
    if (j < 0) {
	pol = nxp_element_mpoly_insert(a, b, 0);
	size = g_slist_length(NXP_ELEMENT_MPOLY(pol)->args->list);
	vector = nxp_element_vector_new(size, NULL);
	zero = nxp_element_int_new(0);
	one = nxp_element_int_new(1);
	one1 = nxp_element_int_new(op == NXP_OP_PLUS ? 1 : -1);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = NXP_ELEMENT_REF(one);
	for (i = 1; i < size; i++)
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i) = NXP_ELEMENT_REF(zero);
	mon = nxp_monomial_new(one1, vector);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(one1);
	NXP_ELEMENT_UNREF(zero);
	NXP_ELEMENT_UNREF(vector);
	nxp_element_list_append(NXP_ELEMENT(NXP_ELEMENT_MPOLY(pol)->mon), mon);
	NXP_ELEMENT_UNREF(mon);
    }
    else {
	size = nxp_element_list_length(NXP_ELEMENT_MPOLY(a)->args);
	vector = nxp_element_vector_new(size, NULL);
	zero = nxp_element_int_new(0);
	one = nxp_element_int_new(1);
	one1 = nxp_element_int_new(op == NXP_OP_PLUS ? 1 : -1);
	for (i = 0; i < size; i++)
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i) = NXP_ELEMENT_REF(j == i ? one : zero);
	mon = nxp_monomial_new(one1, vector);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(one1);
	NXP_ELEMENT_UNREF(zero);
	NXP_ELEMENT_UNREF(vector);
	pol = nxp_element_mpoly_add_monomial(a, mon, context, error);
	NXP_ELEMENT_UNREF(mon);
    }
    return pol;
}

static NxpElement *
sub(NxpOp op, NxpElement *a, NxpElement *b,
	NxpEvalContext *context, GError **error)
{
    NxpElement *pol, *polm;
    NxpElement *vector;
    int i, size, j;
    NxpElement *one, *zero;
    NxpElement *mon;

    g_return_val_if_fail(NXP_ELEMENT_IS_VARIABLE(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_MPOLY(b), NULL);

    j = nxp_element_list_contains(NXP_ELEMENT(NXP_ELEMENT_MPOLY(b)->args), a, context, error);
    if (j < 0) {
	polm = nxp_element_mpoly_insert(b, a, 0);
	pol = nxp_element_mpoly_opposed(polm);
	NXP_ELEMENT_UNREF(polm);
	size = nxp_element_list_length(NXP_ELEMENT(NXP_ELEMENT_MPOLY(pol)->args));
	vector = nxp_element_vector_new(size, NULL);
	zero = nxp_element_int_new(0);
	one = nxp_element_int_new(1);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = NXP_ELEMENT_REF(one);
	for (i = 1; i < size; i++)
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i) = NXP_ELEMENT_REF(zero);
	mon = nxp_monomial_new(one, vector);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(zero);
	NXP_ELEMENT_UNREF(vector);
	nxp_element_list_append(NXP_ELEMENT(NXP_ELEMENT_MPOLY(pol)->mon), mon);
	NXP_ELEMENT_UNREF(mon);
    }
    else {
	size = nxp_element_list_length(NXP_ELEMENT(NXP_ELEMENT_MPOLY(b)->args));
	vector = nxp_element_vector_new(size, NULL);
	zero = nxp_element_int_new(0);
	one = nxp_element_int_new(1);
	for (i = 0; i < size; i++)
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i) = NXP_ELEMENT_REF(j == i ? one : zero);
	mon = nxp_monomial_new(one, vector);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(zero);
	NXP_ELEMENT_UNREF(vector);
	polm = nxp_element_mpoly_opposed(b);
	pol = nxp_element_mpoly_add_monomial(polm, mon, context, error);
	NXP_ELEMENT_UNREF(polm);
	NXP_ELEMENT_UNREF(mon);
    }
    return pol;
}

static NxpElement *
mul(NxpOp op, NxpElement *a, NxpElement *b, 
	NxpEvalContext *context, GError **error)
{
    NxpElement *pol, *retval;

    pol = nxp_typecasts_try_cast(b, NXP_ELEMENT_TYPE_MPOLY);
    g_return_val_if_fail(pol, NULL);

    retval = nxp_element_tree_simplify_op(op, a, pol, context, error);
    NXP_ELEMENT_UNREF(pol);
    return retval;
}

#define BIND_SYM_V(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_VARIABLE, \
   			                              eval, symp,  NXP_OP_BIND_SYMMETRIC, NULL);

#define BIND_P_V(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_VARIABLE, \
   			                              eval, symp,  0, NULL);
#define BIND_V_P(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_VARIABLE, NXP_ELEMENT_TYPE_MPOLY, \
   			                              eval, symp,  0, NULL);

#define BIND_SYM_T(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_TREE, \
   			                              eval, symp,  NXP_OP_BIND_SYMMETRIC, NULL);

#define BIND_P_T(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_MPOLY, NXP_ELEMENT_TYPE_TREE, \
   			                              eval, symp,  0, NULL);
#define BIND_T_P(eval,symp,op)    nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_TREE, NXP_ELEMENT_TYPE_MPOLY, \
   			                              eval, symp,  0, NULL);
void nxp_element_mpoly_variable_operator_init(void)
{
    BIND_SYM_V(NULL, add, NXP_OP_PLUS);
    BIND_SYM_V(NULL, mul, NXP_OP_STAR);
    BIND_P_V(NULL, add, NXP_OP_MINUS);
    BIND_V_P(NULL, sub, NXP_OP_MINUS);
    BIND_SYM_T(NULL, add, NXP_OP_PLUS);
//    BIND_SYM_T(NULL, mul, NXP_OP_STAR);
    BIND_P_T(NULL, add, NXP_OP_MINUS);
    BIND_T_P(NULL, sub, NXP_OP_MINUS);
    BIND_P_V(NULL, mul, NXP_OP_SLASH);
}
