#include "typecasts.h"
#include "vector.h"
#include "types.h"
#include "element_output.h"

// I really don't know where to keep this here or send to another file.
// As this is almost empty and these typecasts are so small, I think there
// isn't a problem to keep them here!
// But, if you have some ideas, mail me!!!!
//
// Albie

NxpElement* matrix2vector (NxpElement *elem) {
    /* This checks if matrix has one dimension equal to 1.  it it
       cant, returns null*/
    int i;
    NxpElement *vector;
    NxpElementMatrix *matrix;

    matrix = NXP_ELEMENT_MATRIX(elem);

    if (matrix->width == 1) {
	vector = nxp_element_vector_new(matrix->height, NULL);
	for (i=0; i<matrix->height; ++i) {
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(vector),i) = 
		NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, 0, i));
	}
    } else {
	if (matrix->height == 1) {
	    vector = nxp_element_vector_new(matrix->width, NULL);
	    for (i=0; i<matrix->width; ++i) {
		NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(vector),i) = 
		    NXP_ELEMENT_REF(NXP_ELEMENT_MATRIX_ELEM_XY(matrix, i, 0));
	    }
	} else {
	    return NULL;
	}
    }
    return vector;
}

NxpElement* r2complex (NxpElement *elem) {
    NxpElement *retval;
    NxpElement *zero = nxp_element_int_new(0);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(elem), NULL);

    retval = nxp_element_complex_new(elem, zero);
    NXP_ELEMENT_UNREF(zero);
    return retval;
}

NxpElement* int2real(NxpElement *elem)
{
    mpfr_t val;
    NxpElement *r;

    mpfr_init_set_z(val, NXP_ELEMENT_INT(elem)->value, GMP_RNDN);

    r = nxp_element_real_new_from_mpfr(val);

    return r;
}

NxpElement* frac2real(NxpElement *elem)
{
    mpfr_t val;
    NxpElement *r;

    mpfr_init_set_q(val, NXP_ELEMENT_FRAC(elem)->value, GMP_RNDN);

    r = nxp_element_real_new_from_mpfr(val);

    return r;
}


NxpElement* variable2mpoly(NxpElement *v)
{
    NxpElement *coeff = nxp_element_int_new(1);
    NxpElement *vector = nxp_element_vector_new(1, NULL);
    NxpElement *retval, *args, *mon;

    _element_vector_set(NXP_ELEMENT_VECTOR(vector), 0, coeff);
    mon = nxp_monomial_new(coeff, vector);
    args = nxp_element_list_new();
    nxp_element_list_append(args, v);
    retval = nxp_element_mpoly_new_simple(mon, args);
    NXP_ELEMENT_UNREF(coeff);
    NXP_ELEMENT_UNREF(vector);
    NXP_ELEMENT_UNREF(args);
    NXP_ELEMENT_UNREF(mon);
    return retval;
}

void nxp_types_init(GError **error)
{
    nxp_element_void_type_init();

    // Number type
    nxp_element_number_type_init();

    // R type
    nxp_element_r_type_init();

    // Basic int type
    nxp_element_int_type_init();

    // Fractions
    nxp_element_frac_type_init();
    nxp_element_frac_int_operator_init();
    nxp_element_frac_real_operator_init();

    // Basic Real type and its relation with int type
    nxp_element_real_type_init();
    nxp_element_real_int_operator_init();

    // Vector type and its relation with previous ones
    nxp_element_vector_type_init();
    nxp_element_vector_any_operator_init();

    // List type and its relation with other ones
    nxp_element_list_type_init();

    // Container type
    nxp_element_container_type_init();

    // Basic Complex type and its relation with int and real type
    nxp_element_complex_type_init();
    nxp_element_complex_r_operator_init();

    // Matrix type and its relation with previous ones
    nxp_element_matrix_type_init();
    nxp_element_matrix_vector_operator_init();
    nxp_element_matrix_number_operator_init();
 
    // String type and its relation with previous ones
    nxp_element_string_type_init();
    nxp_element_string_int_operator_init();

    // Variable type and its relation with previous ones
    nxp_element_variable_type_init();
    nxp_element_variable_number_operator_init();

    // And the reminding structured ones
    nxp_element_function_type_init();
    nxp_element_tree_type_init();

    // Multivariable polynomial
    nxp_element_mpoly_type_init();
    nxp_element_mpoly_variable_operator_init();
    nxp_element_mpoly_number_operator_init();

    // Register TypeCasts
    nxp_typecasts_register_caster( NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_REAL, int2real);
    nxp_typecasts_register_caster( NXP_ELEMENT_TYPE_R, NXP_ELEMENT_TYPE_COMPLEX, r2complex);
    nxp_typecasts_register_caster( NXP_ELEMENT_TYPE_MATRIX, NXP_ELEMENT_TYPE_VECTOR, matrix2vector);
    nxp_typecasts_register_caster( NXP_ELEMENT_TYPE_FRAC, NXP_ELEMENT_TYPE_REAL, frac2real);
    nxp_typecasts_register_caster( NXP_ELEMENT_TYPE_VARIABLE, NXP_ELEMENT_TYPE_MPOLY, variable2mpoly);
}

void nxp_types_function_init(NxpEvalContext *context, GError **error)
{
    nxp_element_mpoly_function_init(context, error);
}
