/* -*- c -*- */
#include <gsl/gsl_sf_trig.h>
#include "nxp_kernel.h"
#include "gsl.h"


// Circular Trignometric Functions
static NxpElement *nxp_gsl_sf_sin(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_sin(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_sf_cos(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_cos(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_sf_hypot(NxpElementReal *arg1 ,NxpElementReal *arg2, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_hypot(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)) ,nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg2)));
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_sf_sinc(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_sinc(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}


NxpElement     *
nxp_gsl_sf_complex_sin_e(NxpElementComplex * Z, ETC)
{
    gsl_sf_result   SZR,
                    SZI;
    gsl_sf_complex_sin_e(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_REAL(Z))),
			 nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_IMG(Z))), &SZR, &SZI);
    GSL_RETURN(nxp_element_complex_new_from_dbl(SZR.val + I * SZI.val));
}

NxpElement     *
nxp_gsl_sf_complex_cos_e(NxpElementComplex * Z, ETC)
{
    gsl_sf_result   SZR,
                    SZI;
    gsl_sf_complex_cos_e(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_REAL(Z))),
			 nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_IMG(Z))), &SZR, &SZI);
    GSL_RETURN(nxp_element_complex_new_from_dbl(SZR.val + I * SZI.val));
}

NxpElement     *
nxp_gsl_sf_complex_logsin_e(NxpElementComplex * Z, ETC)
{
    gsl_sf_result   SZR,
                    SZI;
    gsl_sf_complex_logsin_e(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_REAL(Z))),
			 nxp_element_r_get_value_as_double(NXP_ELEMENT_R(NXP_COMPLEX_IMG(Z))), &SZR, &SZI);
    GSL_RETURN(nxp_element_complex_new_from_dbl(SZR.val + I * SZI.val));
}

static NxpElement *nxp_gsl_sf_lnsinh(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_lnsinh(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_sf_lncosh(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_lncosh(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}


NxpElement     *
nxp_gsl_sf_polar_to_rect(NxpElementVector * V, ETC)
{
    NxpElement     *el_a,
                   *el_b;

    /*
     * We need two values, and exactly two values 
     */
    if (NXP_ELEMENT_VECTOR_SIZE(V) != 2)
	return NULL;
    /*
     * Get the two elements 
     */
    el_a = NXP_ELEMENT_VECTOR_ELEM(V, 0);
    el_b = NXP_ELEMENT_VECTOR_ELEM(V, 1);
    if (NXP_ELEMENT_IS_R(el_a) && NXP_ELEMENT_IS_R(el_b)) {
	NxpElement     *ans;
	double          a = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_a));
	double          b = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_b));
	gsl_sf_result   x,
	                y;
	gsl_sf_polar_to_rect(a, b, &x, &y);
	ans = nxp_element_vector_new(2, NULL);
	_element_vector_index(NXP_ELEMENT_VECTOR(ans), 0) =
	    nxp_element_real_new(x.val);
	_element_vector_index(NXP_ELEMENT_VECTOR(ans), 1) =
	    nxp_element_real_new(y.val);
	GSL_RETURN(ans);
    } else {
	return NULL;
    }
}

NxpElement     *
nxp_gsl_sf_rect_to_polar(NxpElementVector * V, ETC)
{
    NxpElement     *el_a,
                   *el_b;

    /*
     * We need two values, and exactly two values 
     */
    if (NXP_ELEMENT_VECTOR_SIZE(V) != 2)
	return NULL;
    /*
     * Get the two elements 
     */
    el_a = NXP_ELEMENT_VECTOR_ELEM(V, 0);
    el_b = NXP_ELEMENT_VECTOR_ELEM(V, 1);
    if (NXP_ELEMENT_IS_R(el_a) && NXP_ELEMENT_R(el_b)) {
	NxpElement     *ans;
	double          a = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_a));
	double          b = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_b));
	gsl_sf_result   x,
	                y;
	gsl_sf_rect_to_polar(a, b, &x, &y);
	ans = nxp_element_vector_new(2, NULL);
	_element_vector_index(NXP_ELEMENT_VECTOR(ans), 0) =
	    nxp_element_real_new(x.val);
	_element_vector_index(NXP_ELEMENT_VECTOR(ans), 1) =
	    nxp_element_real_new(y.val);
	GSL_RETURN(ans);
    } else {
	return NULL;
    }
}

static NxpElement *nxp_gsl_sf_angle_restrict_symm(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_angle_restrict_symm(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}

static NxpElement *nxp_gsl_sf_angle_restrict_pos(NxpElementReal *arg1, ETC)
{
	nxp_real_t ret;
	ret = gsl_sf_angle_restrict_pos(nxp_element_r_get_value_as_double(NXP_ELEMENT_R(arg1)));
	GSL_RETURN(nxp_element_real_new(ret));
}


// Trigonometric Functions With Error Estimates

void
nxp_gsl_sf_trig_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;
nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"trig");
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "sin");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_sin, NULL, nxp_gsl_sf_sin, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "cos");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_cos, NULL, nxp_gsl_sf_cos, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "hypot");
 nxp_bifunc_define(id, &nxp_prototype__REAL_REAL, nxp_gsl_sf_hypot, NULL, nxp_gsl_sf_hypot, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "sinc");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_sinc, NULL, nxp_gsl_sf_sinc, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "sin");
 nxp_bifunc_define(id, &nxp_prototype__COMPLEX, nxp_gsl_sf_complex_sin_e, NULL, nxp_gsl_sf_complex_sin_e, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "cos");
 nxp_bifunc_define(id, &nxp_prototype__COMPLEX, nxp_gsl_sf_complex_cos_e, NULL, nxp_gsl_sf_complex_cos_e, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "lnSin");
 nxp_bifunc_define(id, &nxp_prototype__COMPLEX, nxp_gsl_sf_complex_logsin_e, NULL, nxp_gsl_sf_complex_logsin_e, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "lnSinh");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_lnsinh, NULL, nxp_gsl_sf_lnsinh, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "lnCosh");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_lncosh, NULL, nxp_gsl_sf_lncosh, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "polar2rect");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_sf_polar_to_rect, NULL, nxp_gsl_sf_polar_to_rect, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "rect2polar");
 nxp_bifunc_define(id, &nxp_prototype__VECTOR, nxp_gsl_sf_rect_to_polar, NULL, nxp_gsl_sf_rect_to_polar, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "restrictSymm");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_angle_restrict_symm, NULL, nxp_gsl_sf_angle_restrict_symm, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","trig", "restrictPos");
 nxp_bifunc_define(id, &nxp_prototype__REAL, nxp_gsl_sf_angle_restrict_pos, NULL, nxp_gsl_sf_angle_restrict_pos, context, NULL); nxp_identifier_unref(id);
}
