#include "nxp_kernel.h"
#include "config.h"

inline static nxp_real_t mycotanl(nxp_real_t x)         { return 1.0 / tanl(x);  }
inline static nxp_complex_t cmycotanl(nxp_complex_t x)  { return 1.0 / ctanl(x); }

inline static nxp_real_t mycotanhl(nxp_real_t x)        { return 1.0 / tanhl(x);  }
inline static nxp_complex_t cmycotanhl(nxp_complex_t x) { return 1.0 / ctanhl(x); }

inline static nxp_real_t mysecl(nxp_real_t x)           { return 1.0 / cosl(x);  }
inline static nxp_complex_t cmysecl(nxp_real_t x)       { return 1.0 / ccosl(x); }

inline static nxp_real_t mycosecl(nxp_real_t x)         { return 1.0 / sinl(x);  }
inline static nxp_complex_t cmycosecl(nxp_real_t x)     { return 1.0 / csinl(x); }

inline static nxp_real_t mysechl(nxp_real_t x)          { return 1.0 / coshl(x);  }
inline static nxp_complex_t cmysechl(nxp_real_t x)      { return 1.0 / ccoshl(x); }

inline static nxp_real_t mycosechl(nxp_real_t x)        { return 1.0 / sinhl(x);  }
inline static nxp_complex_t cmycosechl(nxp_real_t x)    { return 1.0 / csinhl(x); }

inline static nxp_real_t myacotanl(nxp_real_t x)        { return acosl(x)  / asinl(x);  }
inline static nxp_complex_t cmyacotanl(nxp_real_t x)    { return cacosl(x) / casinl(x); }

inline static nxp_real_t myacotanhl(nxp_real_t x)       { return acoshl(x)  / asinhl(x);  }
inline static nxp_complex_t cmyacotanhl(nxp_real_t x)   { return cacoshl(x) / casinhl(x); }

#define EVAL_MPFR(op,func)  NxpElement* stdlib_eval_real_##op(NxpElementReal *el, GError **error) {\
    mpfr_t result;\
    mpfr_init(result);\
    func(result, el->value, GMP_RNDN);\
    return nxp_element_real_new_from_mpfr(result);\
}

#define EVAL(op,func)  NxpElement* stdlib_eval_real_##op(NxpElementR *el, GError **error) {  \
    return nxp_element_real_new(func(nxp_element_r_get_value_as_double(el)));}


#ifdef HAVE_COMPLEX_H
#define EVAC(op,func)  NxpElement* stdlib_eval_complex_##op(NxpElementComplex *el, \
                                                                GError **error) {  \
    return nxp_element_complex_new_from_dbl(func(nxp_element_complex_get_value(el)));}

#else
#define EVAC(op,func)  ;
#endif

EVAL_MPFR(sin, mpfr_sin);
EVAC(sin,csinl);
NxpElement* stdlib_derive_sin(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_TIMES(elem_COS(NXP_ELEMENT_REF(el)), nxp_element_derive(el, var, context, NULL)); 
}
EVAL_MPFR(cos,mpfr_cos);
EVAC(cos,ccosl);
NxpElement* stdlib_derive_cos(NxpIdentifier *var, NxpElement* el, ETC) 
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_SIN(NXP_ELEMENT_REF(el)),
  				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(tan,mpfr_tan);
EVAC(tan,ctanl);
NxpElement* stdlib_derive_tan(NxpIdentifier *var, NxpElement* el, ETC) 
{
    /* TODO: GERROR */
    return elem_TIMES(elem_OVER(nxp_element_real_new(1),
				elem_COS(elem_CIRC(NXP_ELEMENT_REF(el),
						   nxp_element_real_new(2)))),
		      nxp_element_derive(el, var, context, NULL));
}
EVAL_MPFR(cotg, mpfr_cot);
EVAC(cotg, cmycotanl);
NxpElement* stdlib_derive_cotg(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_OVER(nxp_element_real_new(1),
					    elem_SIN(elem_CIRC(NXP_ELEMENT_REF(el),
							       nxp_element_real_new(2)))),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(sinh, mpfr_sinh);
EVAC(sinh, csinhl);
NxpElement* stdlib_derive_sinh(NxpIdentifier *var, NxpElement* el, ETC) 
{
    /* TODO: GERROR */
    return elem_TIMES(elem_COSH(NXP_ELEMENT_REF(el)),
  		      nxp_element_derive(el, var, context, NULL));
}
EVAL_MPFR(cosh, mpfr_cosh);
EVAC(cosh, ccoshl);
NxpElement* stdlib_derive_cosh(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_SINH(NXP_ELEMENT_REF(el)),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(tgh, mpfr_tanh);
EVAC(tgh, ctanhl);
NxpElement* stdlib_derive_tgh(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_TIMES(elem_CIRC(elem_SECH(el),
  				nxp_element_real_new(2)),
  		      nxp_element_derive(el, var, context, NULL));
}
EVAL_MPFR(cotgh, mpfr_coth);
EVAC(cotgh, cmycotanhl);
NxpElement* stdlib_derive_cotgh(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_CIRC(elem_COSECH(el),
					    nxp_element_real_new(2)),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(sec, mpfr_sec);
EVAC(sec, cmysecl);
NxpElement* stdlib_derive_sec(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_TIMES(elem_TIMES(elem_SEC(NXP_ELEMENT_REF(el)),
  				 elem_TG(NXP_ELEMENT_REF(el))),
  		      nxp_element_derive(el, var, context, NULL));
}
EVAL_MPFR(cosec, mpfr_csc);
EVAC(cosec, cmycosecl);
NxpElement* stdlib_derive_cosec(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_TIMES(elem_COSEC(NXP_ELEMENT_REF(el)),
					     elem_COTG(NXP_ELEMENT_REF(el))),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(sech, mpfr_sech);
EVAC(sech, cmysechl);
NxpElement* stdlib_derive_sech(NxpIdentifier *var, NxpElement* el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_TIMES(elem_SECH(NXP_ELEMENT_REF(el)),
					     elem_TGH(NXP_ELEMENT_REF(el))),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(cosech, mpfr_csch);
EVAC(cosech, cmycosechl);
NxpElement* stdlib_derive_cosech(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_TIMES(elem_TIMES(elem_COSECH(NXP_ELEMENT_REF(el)),
					     elem_COTGH(NXP_ELEMENT_REF(el))),
				  nxp_element_derive(el, var, context, NULL)));
}
EVAL_MPFR(arcsin, mpfr_asin);
EVAC(arcsin, casinl);
NxpElement* stdlib_derive_arcsin(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_OVER(nxp_element_derive(el, var, context, NULL),
		     elem_SQRT(elem_MINUS(nxp_element_real_new(1),
					  elem_CIRC(NXP_ELEMENT_REF(el),
						    nxp_element_real_new(2)))));
}
EVAL_MPFR(arccos, mpfr_acos);
EVAC(arccos, cacosl);
NxpElement* stdlib_derive_arccos(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(
  	elem_OVER(nxp_element_derive(el, var, context, NULL),
  		  elem_SQRT(elem_MINUS(nxp_element_real_new(1),
  				       elem_CIRC(NXP_ELEMENT_REF(el), 
  						 nxp_element_real_new(2))))));
}
EVAL_MPFR(arctg, mpfr_atan);
EVAC(arctg, catanl);
NxpElement* stdlib_derive_arctg(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_OVER(nxp_element_derive(el, var, context, NULL),
  		     elem_PLUS(nxp_element_real_new(1), elem_CIRC(NXP_ELEMENT_REF(el),
								  nxp_element_real_new(2))));
}
EVAL(arccotg, myacotanl);
EVAC(arccotg, cmyacotanl);
NxpElement* stdlib_derive_arccotg(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_UMINUS(elem_OVER(nxp_element_derive(el, var, context, NULL),
  				 elem_PLUS(nxp_element_real_new(1),
  					   elem_CIRC(NXP_ELEMENT_REF(el),
  						     nxp_element_real_new(2)))));
}
EVAL_MPFR(arcsinh, mpfr_asinh);
EVAC(arcsinh, casinhl);
NxpElement* stdlib_derive_arcsinh(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_OVER(nxp_element_derive(el, var, context, NULL),
		     elem_SQRT(elem_PLUS(nxp_element_real_new(1),
					 elem_CIRC(NXP_ELEMENT_REF(el),
						   nxp_element_real_new(2)))));
}
EVAL_MPFR(arccosh, mpfr_acosh);
EVAC(arccosh, cacoshl);
NxpElement* stdlib_derive_arccosh(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return
  	elem_OVER(nxp_element_derive(el, var, context, NULL),
  		  elem_TIMES(elem_SQRT(
  		      elem_OVER(elem_MINUS(NXP_ELEMENT_REF(el), nxp_element_real_new(1)),
  				elem_PLUS(NXP_ELEMENT_REF(el), nxp_element_real_new(1)))), 
  			     elem_PLUS(nxp_element_real_new(1),
  				       NXP_ELEMENT_REF(el))));
}
EVAL_MPFR(arctgh, mpfr_atanh);
EVAC(arctgh, catanhl);
NxpElement* stdlib_derive_arctgh(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_OVER(nxp_element_derive(el, var, context, NULL),
  		     elem_MINUS(nxp_element_real_new(1), elem_CIRC(NXP_ELEMENT_REF(el),
								   nxp_element_real_new(2))));
}
EVAL(arccotgh, myacotanhl);
EVAC(arccotgh, cmyacotanhl);
NxpElement* stdlib_derive_arccotgh(NxpIdentifier *var, NxpElement *el, ETC)
{
    /* TODO: GERROR */
    return elem_OVER(nxp_element_derive(el, var, context, NULL),
  		     elem_PLUS(nxp_element_real_new(1), elem_CIRC(NXP_ELEMENT_REF(el),
								  nxp_element_real_new(2))));
}



