#include "vector.h"
#include "i18n.h"

void nxp_stdlib_vector_init(NxpEvalContext *context, GError **error)
{
    GError *err = NULL;

    nxp_bifunc_define_global("reverse", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_reverse, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("swap", &nxp_prototype__VECTOR_INT_INT,
			     nxp_stdlib_vector_swap, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("max", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_max, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("min", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_min, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("max_index", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_max_index, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("min_index", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_min_index, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

    nxp_bifunc_define_global("null", &nxp_prototype__VECTOR,
			     nxp_stdlib_vector_null, NULL, NULL, context, &err);
    if (err) { g_propagate_error(error, err); return;}

}

NxpElement* nxp_stdlib_vector_reverse(NxpElementVector* vector, ETC)
{
    NxpElement *new;
    gint size = NXP_ELEMENT_VECTOR_SIZE(vector), i;
    new = nxp_element_vector_new(size, NULL);
    for (i=0; i<size; ++i) {
	new = nxp_element_vector_real_put(NXP_ELEMENT_VECTOR(new), 
					  NXP_ELEMENT_VECTOR_ELEM(vector, i), 
					  size-1-i); 
    }
    return new;
}

NxpElement* nxp_stdlib_vector_swap(NxpElementVector* vector,
				   NxpElementInt* el_i,
				   NxpElementInt* el_j, ETC) {
    NxpElement* new;
    GError *err = NULL;
    gint i = nxp_element_int_get_si(NXP_ELEMENT(el_i), &err) - 1;
    gint j = nxp_element_int_get_si(NXP_ELEMENT(el_j), &err) - 1;
    gint size = NXP_ELEMENT_VECTOR_SIZE(vector), my_i;

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    if (i<0 || i>=size || j<0 || j>=size) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("index out of range"));
	return NULL;
    }

    new = nxp_element_vector_new(size, NULL);
    for (my_i=0; my_i<size; ++my_i) {
	if (my_i == i) {
	    new = nxp_element_vector_real_put(NXP_ELEMENT_VECTOR(new),
					      NXP_ELEMENT_VECTOR_ELEM(vector, j), i);
	} else if (my_i == j) {
	    new = nxp_element_vector_real_put(NXP_ELEMENT_VECTOR(new),
					      NXP_ELEMENT_VECTOR_ELEM(vector, i), j);
	} else {
	    new = nxp_element_vector_real_put(NXP_ELEMENT_VECTOR(new),
					      NXP_ELEMENT_VECTOR_ELEM(vector, my_i), my_i);
	}
    }
    return new;

}

nxp_real_t nxp_stdlib_real_from_int_or_real(NxpElement* element) {
    return nxp_element_r_get_value_as_double(NXP_ELEMENT_R(element));
}

NxpElement* nxp_stdlib_vector_max(NxpElementVector* vector, ETC) {
    int size = NXP_ELEMENT_VECTOR_SIZE(vector);
    int i;
    if (size < 1) {
	return nxp_element_real_new(0);
    } else {
	nxp_real_t max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, 0));
	for (i=1; i<size; ++i) {
	    nxp_real_t possible_max;
	    possible_max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, i));
	    if (possible_max > max) { max = possible_max; }
	}
	return nxp_element_real_new(max);
    }
}

NxpElement* nxp_stdlib_vector_min(NxpElementVector* vector, ETC) {
    int size = NXP_ELEMENT_VECTOR_SIZE(vector);
    int i;
    if (size < 1) {
	return nxp_element_real_new(0);
    } else {
	nxp_real_t min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, 0));
	for (i=1; i<size; ++i) {
	    nxp_real_t possible_min;
	    possible_min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, i));
	    if (possible_min < min) { min = possible_min; }
	}
	return nxp_element_real_new(min);
    }
}

NxpElement* nxp_stdlib_vector_max_index(NxpElementVector* vector, ETC) {
    int size = NXP_ELEMENT_VECTOR_SIZE(vector);
    int i;
    if (size < 1) {
	return nxp_element_int_new(0);
    } else {
	gint max_index = 0;
	nxp_real_t max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, 0));
	for (i=1; i<size; ++i) {
	    nxp_real_t possible_max;
	    possible_max = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, i));
	    if (possible_max > max) { 
		max = possible_max; 
		max_index = i;
	    }
	}
	return nxp_element_int_new(max_index+1);
    }
}

NxpElement* nxp_stdlib_vector_min_index(NxpElementVector* vector, ETC) {
    int size = NXP_ELEMENT_VECTOR_SIZE(vector);
    int i;
    if (size < 1) {
	return nxp_element_int_new(0);
    } else {
	gint min_index = 0;
	nxp_real_t min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, 0));
	for (i=1; i<size; ++i) {
	    nxp_real_t possible_min;
	    possible_min = nxp_stdlib_real_from_int_or_real(NXP_ELEMENT_VECTOR_ELEM(vector, i));
	    if (possible_min < min) { 
		min = possible_min; 
		min_index = i;
	    }
	}
	return nxp_element_int_new(1+min_index);
    }
}

NxpElement* nxp_stdlib_vector_null(NxpElementVector* vector, ETC) {
    int i;
    int size = NXP_ELEMENT_VECTOR_SIZE(vector);

    for (i=0; i<size; ++i)
	if (!nxp_element_is_true(NXP_ELEMENT_VECTOR_ELEM(vector, i), context, error))
	    return nxp_element_int_new(FALSE);
    return nxp_element_int_new(TRUE);
}

